#lang scheme

;; -- TYPES -- ;;

; node type
; id : number
; outgoing-nodes : listof-number
(define-struct node (id outgoing-nodes))

; edge : (node-id . node-id)

; dfs-state type
; pre-post : hash-from-node-id-to-number-pair
; tree-edges : set-of-edge
; visited : set-of-node-id
; next-num : number
(define-struct dfs-state (pre-post tree-edges visited next-num))

; starting-dfs-state : graph -> dfs-state
(define (starting-dfs-state graph)
  (make-dfs-state (make-immutable-hash
                   (map (lambda (x) (cons x (cons #f #f))) (hash-keys graph)))
                  (make-immutable-hash empty)
                  (make-immutable-hash empty) 
                  1))

; graph : hash-from-node-id-to-node

;; -- Set type and helpers -- ;;

; set a : hash-from-a-to-boolean

; list-to-set : listof-a -> setof-a
(define (list-to-set lst)
  (make-immutable-hash (map (lambda (x) (cons x #t)) lst)))

; set-member : a setof-a -> boolean
(define (set-member e set)
  (hash-has-key? set e))

; set-insert : a setof-a -> setof-a
(define (set-insert e set)
  (hash-set set e #t))
           
; hash-keys : hash-from-a-to-? -> listof-a
(define (hash-keys hash) (for/list ([x (in-hash-keys hash)]) x))

;; -- DFS -- ;;

; dfs : graph -> dfs-state
; dfs pseudo-code:
; dfs graph =
;   for-each node in graph
;     visit node
(define (dfs graph)
  (dfs-iterate graph (hash-keys graph) (starting-dfs-state graph)))

; dfs-iterate : graph listof-node-id dfs-state -> dfs-state
(define (dfs-iterate graph nodes-to-visit s)
  (cond
    [(empty? nodes-to-visit) s]
    [else (let ((s1 (visit (first nodes-to-visit) #f graph s)))
            (dfs-iterate graph (rest nodes-to-visit) s1))]))

; visit : node-id node-id-or-false graph dfs-state -> dfs-state
; visit pseudo-code:
; visit node-id came-from graph =
;  unless already-visited node then
;    mark-visited node-id
;    if came-from then add-tree-edge (came-from . node-id)
;    mark-pre-number node-id
;    for child-id in node-children node:
;      visit child-id node-id graph
;    mark-post-number node-id
(define (visit node-id came-from graph s)
  (cond
    [(set-member node-id (dfs-state-visited s)) s]
    [else (let* ((s1-visited (set-insert node-id (dfs-state-visited s)))
                 (s1-tree-edges (if (not came-from)
                                    (dfs-state-tree-edges s)
                                    (set-insert (cons came-from node-id) (dfs-state-tree-edges s))))
                 (s1-pre-post (hash-update (dfs-state-pre-post s) 
                                           node-id 
                                           (lambda (x) (cons (dfs-state-next-num s) (cdr x)))))
                 (s1-next-num (+ 1 (dfs-state-next-num s)))
                 (s1 (make-dfs-state s1-pre-post s1-tree-edges s1-visited s1-next-num))
                 (s2 (visit-many (node-outgoing-nodes (hash-ref graph node-id)) node-id graph s1))
                 (s3-pre-post (hash-update (dfs-state-pre-post s2)
                                           node-id
                                           (lambda (x) (cons (car x) (dfs-state-next-num s2)))))
                 (s3-next-num (+ 1 (dfs-state-next-num s2)))
                 (s3 (make-dfs-state s3-pre-post (dfs-state-tree-edges s2) (dfs-state-visited s2) s3-next-num)))
            s3)]))

; visit-many : list-of-node-id node-id-or-false graph dfs-state -> dfs-state
(define (visit-many node-ids came-from graph s)
  (cond
    [(empty? node-ids) s]
    [else (let ((s1 (visit (first node-ids) came-from graph s)))
            (visit-many (rest node-ids) came-from graph s1))]))

               
;; -- PARSING HELPERS -- ;;

(define (make-graph sexp)
  (cond
    [(empty? sexp) (make-immutable-hash empty)]
    [else (let* ((unparsed-node (first sexp))
                 (node-id (first unparsed-node))
                 (node-outgoing (rest unparsed-node))
                 (node (make-node node-id node-outgoing)))
            (hash-set (make-graph (rest sexp)) node-id node))]))

;; -- TESTING HELPERS -- ;;

(define (test a b)
  (if (equal? a b) 
      (void)
      (error "test failed")))

;; -- TESTS -- ;;

(define graph1 (make-graph '((1 2 3) (2 3) (3))))
(test (dfs-state-pre-post (dfs graph1)) 
      #hash((1 . (1 . 6))
            (2 . (2 . 5)) 
            (3 . (3 . 4))))
(test (dfs-state-tree-edges (dfs graph1))
      #hash(((1 . 2) . #t)
            ((2 . 3) . #t)))

(define graph2 (make-graph '((1 2 3) (2) (3) (4 5 6 7) (5 7) (6 7) (7))))
(test (dfs-state-pre-post (dfs graph2)) 
      #hash((1 . (1 . 6)) 
            (2 . (2 . 3))
            (3 . (4 . 5)) 
            (4 . (7 . 14)) 
            (5 . (8 . 11))
            (6 . (12 . 13))
            (7 . (9 . 10))))
(test (dfs-state-tree-edges (dfs graph2))
      #hash(((1 . 2) . #t)
            ((1 . 3) . #t)
            ((4 . 5) . #t)
            ((4 . 6) . #t)
            ((5 . 7) . #t)))