#lang scheme

; Monad type State s a = (s -> (a . s))
; or SymbolState a = (symbol -> (a . symbol))

; return : a -> (m a)
; return : a -> (s -> (a . s))
(define (return a)
  (lambda (s) (cons a s)))

; >>= : (m a)          (a -> (m b))          -> (m b)
; >>= : (s -> (a . s)) (a -> (s -> (b . s))) -> (s -> (b . s))
(define (>>= sM f)
  (lambda (s)
    (let* ((a*s1 (sM s))
           (a (car a*s1))
           (s1 (cdr a*s1)))
      ((f a) s1))))

; >> : (m a)          (m b)          -> (m b)
; >> : (s -> (a . s)) (s -> (b . s)) -> (s -> (b . s))
(define (>> aM bM)
  (>>= aM (lambda (_) bM)))

; put : a -> (m ())
; put : a -> (a -> (() . a))
(define (put a)
  (lambda (x) (cons #f a)))

; get : (m a)
; get : a -> (a . a)
(define get
  (lambda (x) (cons x x)))

; modify : (a -> a) -> (m ())
(define (modify f)
  (>>= get (lambda (v) (put (f v)))))

; eval-state : (s -> (a . s)) s -> a
(define (eval-state f s)
  (car (f s)))

; exec-state : (s -> (a . s)) s -> s
(define (exec-state f s)
  (cdr (f s)))

; for-each-M : list-of-a (a -> m ?_1) -> m ?_2
; where ?_2 is the last ?_1 or #f if list is empty
(define (for-each-M list f)
  (foldl (lambda (x m) (>> m (f x))) (return #f) list))

; when : boolean (m ()) -> m ()
(define (when b action)
  (if b action (return #f)))

; unless : boolean (m ()) -> m ()
(define (unless b action)
  (if b (return #f) action ))

(define-syntax letM
  (syntax-rules ()
    ((_ (v monad-e) result-e)
     (>>= monad-e (lambda (v) result-e)))))

(define-syntax do
  (syntax-rules (<-)
    ((_ (<- v monad-e) . rest)
     (letM (v monad-e) (do . rest)))
    ((_ monad-e)
     monad-e)
    ((_ monad-e . rest)
     (>> monad-e (do . rest)))))

(provide return >>= >> put get modify eval-state exec-state for-each-M when unless letM do)
